import logging
from json.decoder import JSONDecodeError
import pandas as pd
import numpy as np
from .EuroLeagueData import EuroLeagueData
from .boxscore_data import BoxScoreData
from .utils import get_requests

logger = logging.getLogger(__name__)


class PlayByPlay(EuroLeagueData):
    """
    A class for getting the game play-by-play data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_game_play_by_play_data(
        self,
        season: int,
        gamecode: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of a particular game.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of the game.
        """
        url = "https://live.euroleague.net/api/PlaybyPlay"
        params = {
            "gamecode": gamecode,
            "seasoncode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)

        try:
            data = r.json()
        except JSONDecodeError as exc:
            raise ValueError(
                f"Game code, {gamecode}, season {season}, "
                "did not return any data."
            ) from exc

        periods = [
            'FirstQuarter', 'SecondQuarter', 'ThirdQuarter', 'ForthQuarter',
            'ExtraTime'
        ]
        all_data = []
        for p, period in enumerate(periods):
            if data[period]:
                df = pd.json_normalize(data[period])
                df["PERIOD"] = p + 1
                all_data.append(df)

        pbp_df = pd.concat(all_data).reset_index(drop=True)
        pbp_df['CODETEAM'] = pbp_df['CODETEAM'].str.strip()
        pbp_df['PLAYER_ID'] = pbp_df['PLAYER_ID'].str.strip()
        pbp_df["PLAYTYPE"] = pbp_df["PLAYTYPE"].str.strip()
        pbp_df["MARKERTIME"] = pbp_df["MARKERTIME"].str.strip()
        pbp_df.insert(0, 'Season', season)
        pbp_df.insert(1, 'Gamecode', gamecode)
        return pbp_df

    def get_play_by_play_data_round(
        self,
        season: int,
        round_number: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a single
        round

        Args:
            season (int): The start year of the season
            round_number (int): The round of the season

        Returns:
            pd.DataFrame: A dataframe with the play-by-play data of all games
                in a single round_number
        """
        df = self.get_round_data_from_game_data(
            season, round_number, self.get_game_play_by_play_data)
        return df

    def get_game_play_by_play_data_single_season(
        self,
        season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a single
        season

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in a single season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_play_by_play_data)
        return data_df

    def get_game_play_by_play_data_multiple_seasons(
        self, start_season: int, end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a range of
        seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in range of seasons
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_play_by_play_data)
        return df

    def get_pbp_data_with_lineups(
        self,
        season,
        gamecode,
        validate=True
    ) -> pd.DataFrame:
        """
        Get the play-by-play (PBP) data enriched with the teams' lineups for
        every action in the PBP data.

        There are three cases where the player in the corresponding row
        (action) is not part of the lineup:

            1. When the player is subbed "OUT", the assigned lineup does not
            contain player name. Since this is reasonable, the value of the
            `validate_on_court_player` indicator is set to `True`.

            2. A player passes on to a teamate, who draws a shooting foul. If
            the passer is subbed before the first free throw and the first free
            throw is made then the subbed player is given an assist. However,
            the player is not in the extracted lineup because he has already
            been subbed. The value of the `validate_on_court_player`
            indicator is set to `False`. We don't fix the lineup, because it
            breaks the lineup continuinity. It is a quirk of the data
            collection andrecording.

            3. There are a few instances where a sub is recorded many seconds
            and actions since it actually happened. This causes issues, such
            as a player records an actions, such as a rebound, but he comes
            on the court (according to the data) many seconds after. This has
            been validated by watching theactual footage of the game. This
            requires a lot of manual work, which is beyond the score of this
            library. It is another quirk of the datacollection and recording,
            hence the value of the `validate_on_court_player` indicator is set
            to `False`.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

            validate (bool, optional): A bool indicator whether to enrich the
                dataframe with two extra columns, which validate the validity
                and consistency of the extracted lineup. Defaults to True.

        Returns:

            pd.DataFrame: A dataframe with the play-by-play enriched with
                teams' lineups
        """
        def process_sub(five, player, sub_type):
            opp_sub_type = "OUT" if sub_type == "IN" else "IN"
            potential_indx = (
                pbp_data.loc[idx + 1:].index[
                    (pbp_data.loc[idx + 1:, "PLAYTYPE"] == opp_sub_type)
                ]
            )
            potential_nops = (
                set(potential_indx).difference(processed_idxs)
            )
            matching_idx = min(potential_nops)
            processed_idxs.append(idx)
            processed_idxs.append(matching_idx)
            matching_row = pbp_data.loc[matching_idx]
            invalid_mask = (
                (matching_row["PLAYTYPE"] != opp_sub_type) or
                (matching_row["CODETEAM"] != team) or
                (matching_row["MARKERTIME"] != markertime)
            )
            if invalid_mask:
                logger.warning(
                    f"Something went wrong for gamecode {gamecode} at sub "
                    f"index {idx} with matching sub index {matching_idx}"
                )
            player_sub = matching_row["PLAYER"].replace(" ,", ",")
            player_in = player if sub_type == "IN" else player_sub
            player_out = player_sub if sub_type == "IN" else player
            if player_in == player_out:
                # there are instance where the same player is subbed in and out
                return five
            else:
                pindx = five.index(player_out)
                five = five[:pindx] + [player_in] + five[pindx + 1:]
                return five

        def validate_player(x, col1, col2):
            flag = False
            if (x["PLAYER"] is not None) and (x["PLAYTYPE"] != "OUT"):
                if x["PLAYER"].replace(" ,", ",") in (x[col1] + x[col2]):
                    flag = True
            else:
                flag = True
            return flag

        # Get the starting line-ups from boxscore data
        boxscoredata = BoxScoreData(competition=self.competition)
        game_bxscr_stats = boxscoredata.get_player_boxscore_stats_data(
            season=season, gamecode=gamecode)

        # find home and away teams
        hm_aw = game_bxscr_stats[["Home", "Team"]].drop_duplicates()
        home_team = hm_aw.loc[hm_aw["Home"] == 1, "Team"].values[0]
        away_team = hm_aw.loc[hm_aw["Home"] == 0, "Team"].values[0]

        starting_five = game_bxscr_stats.loc[
            game_bxscr_stats["IsStarter"] == 1, ["Team", "Player"]
        ]
        starting_five['ID'] = starting_five.groupby('Team').cumcount()

        # Pivot the DataFrame
        starting_five = starting_five.pivot(
            index='ID', columns='Team', values='Player')

        # Reset index if needed
        starting_five.reset_index(drop=True, inplace=True)
        starting_five = starting_five[[home_team, away_team]]
        starting_five_dict = starting_five.to_dict(orient='list')

        # Fetch play-by-play data
        pbp_data = self.get_game_play_by_play_data(
            season=season, gamecode=gamecode)

        # Asign the starting lineups to the first entry of the PBP data.
        pbp_data["Lineup_A"] = None
        pbp_data["Lineup_B"] = None
        pbp_data.at[0, "Lineup_A"] = starting_five_dict[home_team]
        pbp_data.at[0, "Lineup_B"] = starting_five_dict[away_team]
        pbp_data["IsHomeTeam"] = np.where(
            pbp_data["CODETEAM"] == home_team, True,
            np.where(pbp_data["CODETEAM"] == away_team,
                     False, None)  # type: ignore
        )

        # start processing the sub entries, row by row.
        processed_idxs: list = []
        current_five_home = starting_five_dict[home_team].copy()
        current_five_away = starting_five_dict[away_team].copy()
        for idx, row in pbp_data.iterrows():
            playtype = row["PLAYTYPE"]
            # there are rare instances of an extra space in player name
            player = (
                row["PLAYER"] if row["PLAYER"] is None
                else row["PLAYER"].replace(" ,", ",")
            )
            team = row["CODETEAM"]
            markertime = row["MARKERTIME"]
            if team != "":
                is_home = home_team == team
                five = current_five_home if is_home else current_five_away
                if (playtype == "OUT") and (idx not in processed_idxs):
                    five = process_sub(five, player, playtype)
                elif (playtype == "IN") and (idx not in processed_idxs):
                    five = process_sub(five, player, playtype)
                if is_home:
                    current_five_home = five
                else:
                    current_five_away = five

            pbp_data.at[idx, "Lineup_A"] = current_five_home
            pbp_data.at[idx, "Lineup_B"] = current_five_away

        if validate:
            lu_cols = [u for u in pbp_data.columns if u.startswith("Lineup_")]
            pbp_data["validate_on_court_player"] = pbp_data.apply(
                lambda x: validate_player(x, lu_cols[0], lu_cols[1]),
                axis=1
            )
        return pbp_data

    def get_pbp_data_with_lineups_round(
        self,
        season: int,
        round_number: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play with lineups data of *all* games
        in a single round

        Args:
            season (int): The start year of the season
            round_number (int): The round of the season

        Returns:
            pd.DataFrame: A dataframe with the play-by-play data with lineups
                of all games in a single round
        """
        df = self.get_round_data_from_game_data(
            season, round_number, self.get_pbp_data_with_lineups)
        return df

    def get_pbp_data_with_lineups_single_season(
        self,
        season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data enriched with team lineups
        of *all* games in a single season

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in a single season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_pbp_data_with_lineups)
        return data_df

    def get_pbp_data_with_lineups_multiple_seasons(
        self, start_season: int, end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data enriched with team lineups
        of *all* games in a range of seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in range of seasons
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_pbp_data_with_lineups)
        return df



from typing import Optional
import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .utils import (
    raise_error,
    get_requests
)


class PlayerStats(EuroLeagueData):
    """
    A class for getting the player-level stats and data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_player_stats(
        self,
        endpoint: str,
        params: dict = {},
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A wrapper function for getting the players' stats for
        - all seasons
        - a single season
        - a range of seasons

        Args:

            endpoint (str): The type of stats, available variables:
                - traditional
                - advanced
                - misc
                - scoring

            params (Dict[str, Union[str, int]]): A dictionary of parameters
                for the get request.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                Defaults to "PerGame".

        Raises:

            ValueError: If the endpoint is not applicable

            ValueError: If the phase_type_code is not applicable

            ValueError: If the statistic_mode is not applicable

        Returns:

            pd.DataFrame: A dataframe with the players' stats.
        """

        available_endpoints = ["traditional", "advanced", "misc", "scoring"]
        available_phase_type_code = ["RS", "PO", "FF"]
        available_stat_mode = ["PerGame", "Accumulated", "Per100Possesions"]

        raise_error(
            endpoint, "Statistic type", available_endpoints, False)
        raise_error(
            statistic_mode, "Statistic Aggregation", available_stat_mode,
            False)
        raise_error(
            phase_type_code, "Phase type code", available_phase_type_code,
            True)

        params["statisticMode"] = statistic_mode
        params["phaseTypeCode"] = phase_type_code
        params["limit"] = 400

        url_ = f"{self.url}/statistics/players/{endpoint}"

        r = get_requests(url_, params=params)
        data = r.json()
        if data["total"] > len(data["players"]):
            params["limit"] = data["total"] + 1
            r = get_requests(url_, params=params)
            data = r.json()
        df = pd.json_normalize(data["players"])
        return df

    def get_player_stats_leaders(
        self,
        params: dict = {},
        stat_category: str = "Score",
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None,
        position: Optional[str] = None
    ) -> pd.DataFrame:
        """
        A wrapper function for collecting the leading players in a given
        stat category.

        Args:

            params (Dict[str, Union[str, int]]): A dictionary of parameters
                for the get request.

            top_n (int): The number of top N players to return.
                Defaults to 200.

            stat_category (str): The stat category. Available values:
                - Valuation
                - Score
                - TotalRebounds
                - OffensiveRebounds
                - Assistances
                - Steals
                - BlocksFavour
                - BlocksAgainst
                - Turnovers
                - FoulsReceived
                - FoulsCommited
                - FreeThrowsMade
                - FreeThrowsAttempted
                - FreeThrowsPercent
                - FieldGoalsMade2
                - FieldGoalsAttempted2
                - FieldGoals2Percent
                - FieldGoalsMade3
                - FieldGoalsAttempted3
                - FieldGoals3Percent
                - FieldGoalsMadeTotal
                - FieldGoalsAttemptedTotal
                - FieldGoalsPercent
                - AccuracyMade
                - AccuracyAttempted
                - AccuracyPercent
                - AssitancesTurnoversRation
                - GamesPlayed
                - GamesStarted
                - TimePlayed
                - Contras
                - Dunks
                - OffensiveReboundPercentage
                - DefensiveReboundPercentage
                - ReboundPercentage
                - EffectiveFeildGoalPercentage
                - TrueShootingPercentage
                - AssistRatio
                - TurnoverRatio
                - FieldGoals2AttemptedRatio
                - FieldGoals3AttemptedRatio
                - FreeThrowRate
                - Possessions
                - GamesWon
                - GamesLost
                - DoubleDoubles
                - TripleDoubles
                - FieldGoalsAttempted2Share
                - FieldGoalsAttempted3Share
                - FreeThrowsAttemptedShare
                - FieldGoalsMade2Share
                - FieldGoalsMade3Share
                - FreeThrowsMadeShare
                - PointsMade2Rate
                - PointsMade3Rate
                - PointsMadeFreeThrowsRate
                - PointsAttempted2Rate
                - PointsAttempted3Rate
                - Age

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

            position (Optional[str], optional): The position of the player to
                draw the top stats from. Available values:
                - Guards
                - Forwards
                - Centers
                - RisingStars
                Defaults to None, meaning all positions.

        Raises:

            ValueError: If the stat_category is not applicable

            ValueError: If the phase_type_code is not applicable

            ValueError: If the statistic_mode is not applicable

            ValueError: If the game_type is not applicable

            ValueError: If the position is not applicable

        Returns:

            pd.DataFrame: A dataframe with the top players' stats
        """
        avaiable_stat_category = [
            "Valuation",
            "Score",
            "TotalRebounds",
            "OffensiveRebounds",
            "Assistances",
            "Steals",
            "BlocksFavour",
            "BlocksAgainst",
            "Turnovers",
            "FoulsReceived",
            "FoulsCommited",
            "FreeThrowsMade",
            "FreeThrowsAttempted",
            "FreeThrowsPercent",
            "FieldGoalsMade2",
            "FieldGoalsAttempted2",
            "FieldGoals2Percent",
            "FieldGoalsMade3",
            "FieldGoalsAttempted3",
            "FieldGoals3Percent",
            "FieldGoalsMadeTotal",
            "FieldGoalsAttemptedTotal",
            "FieldGoalsPercent",
            "AccuracyMade",
            "AccuracyAttempted",
            "AccuracyPercent",
            "AssitancesTurnoversRation",
            "GamesPlayed",
            "GamesStarted",
            "TimePlayed",
            "Contras",
            "Dunks",
            "OffensiveReboundPercentage",
            "DefensiveReboundPercentage",
            "ReboundPercentage",
            "EffectiveFeildGoalPercentage",
            "TrueShootingPercentage",
            "AssistRatio",
            "TurnoverRatio",
            "FieldGoals2AttemptedRatio",
            "FieldGoals3AttemptedRatio",
            "FreeThrowRate",
            "Possessions",
            "GamesWon",
            "GamesLost",
            "DoubleDoubles",
            "TripleDoubles",
            "FieldGoalsAttempted2Share",
            "FieldGoalsAttempted3Share",
            "FreeThrowsAttemptedShare",
            "FieldGoalsMade2Share",
            "FieldGoalsMade3Share",
            "FreeThrowsMadeShare",
            "PointsMade2Rate",
            "PointsMade3Rate",
            "PointsMadeFreeThrowsRate",
            "PointsAttempted2Rate",
            "PointsAttempted3Rate",
            "Age"
        ]
        available_phase_type_code = ["RS", "PO", "FF"]
        available_stat_mode = [
            "PerGame",
            "Accumulated",
            "Per100Possesions",
            "PerGameReverse",
            "AccumulatedReverse"
        ]
        available_game_types = [
            "HomeGames",
            "AwayGames",
            "GamesWon",
            "GamesLost",
        ]
        availabe_positions = [
            "Guards",
            "Forwards",
            "Centers",
            "RisingStars"
        ]

        raise_error(stat_category, "Stat category",
                    avaiable_stat_category, False)
        raise_error(
            statistic_mode, "Statistic Aggregation", available_stat_mode,
            False)
        raise_error(
            phase_type_code, "Phase type code", available_phase_type_code,
            True)
        raise_error(game_type, "Game type", available_game_types, True)
        raise_error(position, "Position", availabe_positions, True)

        params["category"] = stat_category
        params["phaseTypeCode"] = phase_type_code
        params["statisticMode"] = statistic_mode
        params["limit"] = top_n
        if game_type is not None and position is not None:
            raise ValueError(
                "Cannot select a game_type and position at the same type. "
                "One of the variables must be None"
            )
        elif game_type is not None and position is None:
            params["misc"] = game_type
        elif game_type is None and position is not None:
            params["misc"] = position

        url_ = f"{self.url}/stats/players/leaders"

        r = get_requests(url_, params=params)
        data = r.json()
        df = pd.json_normalize(data["data"])
        return df

    def get_player_stats_all_seasons(
        self,
        endpoint: str,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        The players' stats for *all* seasons.

        Args:

            endpoint (str): The type of stats, available variables:
                - traditional
                - advanced
                - misc
                - scoring

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the players' stats
        """
        params = {"SeasonMode": "All"}
        df = self.get_player_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_player_stats_single_season(
        self,
        endpoint: str,
        season: int,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        The players' stats for a *single* season.

        Args:

            endpoint (str): The type of stats, available variables:
                - traditional
                - advanced
                - misc
                - scoring

            season (int): The start year of the season.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the players' stats
        """
        params = {
            "SeasonMode": "Single",
            "SeasonCode": f"{self.competition}{season}",
        }
        df = self.get_player_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_player_stats_range_seasons(
        self,
        endpoint: str,
        start_season: int,
        end_season: int,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        The players' stats for a range of seasons.

        Args:

            endpoint (str): The type of stats, available variables:
                - traditional
                - advanced
                - misc
                - scoring

            start_season (int): The start year of the first season in the
                range.

            end_season (int): The start year of teh last season in the range.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the players' stats
        """
        params = {
            "SeasonMode": "Range",
            "FromSeasonCode": f"{self.competition}{start_season}",
            "ToSeasonCode": f"{self.competition}{end_season}",
        }
        df = self.get_player_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_player_stats_leaders_all_seasons(
        self,
        stat_category: str,
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None,
        position: Optional[str] = None
    ) -> pd.DataFrame:
        """
        Get the top leaders in a statistical category in all seasons

        Args:

            stat_category (str): The stat category. See function
                `utils.get_player_stats_leaders` for a list of available stats.

            top_n (int): The number of top N players to return.
                Defaults to 200.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

            position (Optional[str], optional): The position of the player to
                draw the top stats from. Available values:
                - Guards
                - Forwards
                - Centers
                - RisingStars
                Defaults to None, meaning all positions.

        Returns:

            pd.DataFrame: A dataframe with the top leading players and their
                stat
        """
        params = {"SeasonMode": "All"}
        df = self.get_player_stats_leaders(
            params,
            stat_category,
            top_n,
            phase_type_code,
            statistic_mode,
            game_type,
            position
        )
        return df

    def get_player_stats_leaders_single_season(
        self,
        season: int,
        stat_category: str,
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None,
        position: Optional[str] = None
    ) -> pd.DataFrame:
        """
        Get the top leaders in a statistical category in a single season

        Args:

            season (int): The start year of the season.

            stat_category (str): The stat category. See function
                `utils.get_player_stats_leaders` for a list of available stats.
            top_n (int): The number of top N players to return.
                Defaults to 200.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

            position (Optional[str], optional): The position of the player to
                draw the top stats from. Available values:
                - Guards
                - Forwards
                - Centers
                - RisingStars
                Defaults to None, meaning all positions.

        Returns:

            pd.DataFrame: A dataframe with the top leading players and their
                stat
        """
        params = {
            "SeasonMode": "Single",
            "SeasonCode": f"{self.competition}{season}",
        }
        df = self.get_player_stats_leaders(
            params,
            stat_category,
            top_n,
            phase_type_code,
            statistic_mode,
            game_type,
            position
        )
        return df

    def get_player_stats_leaders_range_seasons(
        self,
        start_season: int,
        end_season: int,
        stat_category: str,
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None,
        position: Optional[str] = None
    ) -> pd.DataFrame:
        """
        Get the top leaders in a statistical category in a range of seasons

        Args:

            start_season (int): The start year of the first season in the
                range.

            stat_category (str): The stat category. See function
                `utils.get_player_stats_leaders` for a list of available stats.

            top_n (int): The number of top N players to return.
                Defaults to 200.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

            position (Optional[str], optional): The position of the player to
                draw the top stats from. Available values:
                - Guards
                - Forwards
                - Centers
                - RisingStars
                Defaults to None, meaning all positions.

        Returns:

            pd.DataFrame: A dataframe with the top leading players and their
                stat
        """
        params = {
            "SeasonMode": "Range",
            "FromSeasonCode": f"{self.competition}{start_season}",
            "ToSeasonCode": f"{self.competition}{end_season}",
        }
        df = self.get_player_stats_leaders(
            params,
            stat_category,
            top_n,
            phase_type_code,
            statistic_mode,
            game_type,
            position
        )
        return df



from typing import Optional, List, Callable
import requests
from requests.exceptions import HTTPError
import logging
import pandas as pd
from tqdm.auto import tqdm

logging.basicConfig(encoding='utf-8', level=logging.INFO)
logger = logging.getLogger(__name__)


def get_requests(
    url: str,
    params: dict = {},
    headers: dict = {"Accept": "application/json"}
) -> requests.models.Response:
    """
    A wrapper to `requests.get()` which handles unsuccesful requests too.

    Args:

        url (str): _description_

        params (dict, optional): The `params` variables in get requests.
            Defaults to {}.

        headers (dict, optional): the `header` variable in get requests.
            Defaults to {"Accept": "application/json"}.

    Raises:

        Requests Error: If get request was not succesful

    Returns:

        requests.models.Response: The response object.
    """
    r = requests.get(url, params=params, headers=headers, timeout=60)

    if r.status_code != 200:
        r.raise_for_status()

    return r


def raise_error(
    var: Optional[str],
    descripitve_var: str,
    available_vals: List,
    allow_none: bool = False,
) -> None:
    """
    A function that raises a ValueError with specific message.

    Args:

        var (str): The input variable by the user

        descripitve_var (str): A description in plain English of this variable

        available_vals (List): The available variables

        allow_none (bool, optional): If `var` can take None value.
            Defaults to False.

    Raises:

        ValueError: if `var` not applicable
    """

    if allow_none:
        available_vals.append(None)

    if var not in available_vals:
        raise ValueError(
            f"{descripitve_var}, {var}, is not applicable. "
            f"Available values: {available_vals}"
        )
    return


def get_data_over_collection_of_games(
    game_codes_df,
    season: int,
    fun: Callable[[int, int], pd.DataFrame]
) -> pd.DataFrame:
    """A function that collects data over a collection of games given their
    game codes. It is a wrapper function that calls the `fun` function

    Args:
        game_codes_df (_type_): _description_
        season (int, optional): The start year of the season.
        fun (Callable[[int, int], pd.DataFrame]): A callable function that
            determines that type of data to be collected. Available values:
            - get_game_report
            - get_game_stats
            - get_game_teams_comparison
            - get_game_play_by_play_data
            - get_game_shot_data
            - get_game_boxscore_quarter_data
            - get_player_boxscore_stats_data
            - get_game_metadata


    Returns:
        pd.DataFrame: A dataframe with the corresponding data of all
            games in the collection.
    """
    data_list = []
    for _, row in tqdm(game_codes_df.iterrows(),
                       total=game_codes_df.shape[0],
                       desc=f"Season {season}", leave=True):
        game_code = row["gameCode"]
        try:
            df = fun(season, game_code)
            if df.empty:
                logger.warning(
                    f"Game {game_code}, season {season} returned no data."
                )
                continue
            if ("Phase" not in df.columns) and ("Phase" in row):
                df.insert(1, "Phase", row["Phase"])
            if ("Round" not in df.columns) and ("Round" in row):
                df.insert(2, "Round", row["Round"])
            data_list.append(df)
        except HTTPError as err:
            logger.warning(
                f"HTTPError: Didn't find gamecode {game_code} for season "
                f"{season}. Invalid {err}. Skip and continue."
            )
        except Exception as e:  # noqa: E722
            logger.warning(
                f"\nSomething went wrong for game {game_code}, "
                f"season {season}.\nError message: {e}. "
                "\nSkip and continue"
            )

    if data_list:
        data_df = pd.concat(data_list, axis=0)
        data_df.reset_index(drop=True, inplace=True)
    else:
        data_df = pd.DataFrame([])
    return data_df



from json.decoder import JSONDecodeError
import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .utils import get_requests


class GameMetadata(EuroLeagueData):

    """
    A class for getting the game related metadata, such as
    stadum, capacity and referee names.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_game_metadata(self, season: int, gamecode: int) -> pd.DataFrame:
        """
        Retrieves game metadata for a given gamecode, such as stadium,
        capacity and referee names.

        Args:
            season (int): The start year of the season.
            gamecode (int): Unique identifier code of the game which can be
                found on Euroleague's official website.

        Returns:
            pd.DataFrame: A dataframe containing metadata of a game.
        """
        url = "https://live.euroleague.net/api/Header"
        params = {
            "gamecode": gamecode,
            "seasoncode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)

        try:
            data = r.json()
        except JSONDecodeError as exc:
            raise ValueError(
                f"Game code, {gamecode}, season {season} "
                "did not return any data."
            ) from exc
        metadata_df = pd.json_normalize(data)
        metadata_df.insert(0, 'Season', season)
        metadata_df.insert(1, 'Gamecode', gamecode)
        metadata_df["Round"] = metadata_df["Round"].astype(int)
        return metadata_df

    def get_game_metadata_round(
        self, season: int, round_number: int
    ) -> pd.DataFrame:
        """
        A function that gets the metadata of all games in a single round.

        Args:
            season (int): The start year of the season.
            round_number (int): The round of the season.

        Returns:
            pd.DataFrame: A dataframe with the metadata of all games in a
                single round.
        """
        df = self.get_round_data_from_game_data(
            season=season,
            round_number=round_number,
            fun=self.get_game_metadata
        )
        return df

    def get_game_metadata_single_season(self, season: int) -> pd.DataFrame:
        """
        A function to retrieve game metadata for all games in a single season.

        Args:
            season (int): The start year of the season.

        Returns:
            pd.DataFrame: A dataframe containing metadata for all games
                in the given season.
        """

        single_season_metadata_df = self.get_season_data_from_game_data(
            season, self.get_game_metadata
        )
        return single_season_metadata_df

    def get_game_metadata_multiple_seasons(
        self, start_season: int, end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the metadata of *all* games in a range of seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the metadata of all games in range
                of seasons
        """
        metadata_df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_metadata)
        return metadata_df



from typing import Callable
import logging
from json.decoder import JSONDecodeError
import pandas as pd
from tqdm.auto import trange
import xmltodict
from .utils import get_requests, get_data_over_collection_of_games

logging.basicConfig(encoding='utf-8', level=logging.INFO)
logger = logging.getLogger(__name__)


class EuroLeagueData:
    """
    Base class for collecting Euroleague and Eurocup competition's data.

    Args:
        competition (str, optional): The competition code. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """
    BASE_URL = "https://api-live.euroleague.net"
    V3 = "v3"
    V2 = "v2"
    V1 = "v1"

    def __init__(self, competition="E"):
        """init function for the EuroLeagueData class.

        Args:
            competition (str, optional): The competition code. Choose one of:
                - 'E' for Euroleague
                - 'U' for Eurocup
                Defaults to "E".

        Raises:
            ValueError: When an invalid competition code is provided.
        """
        if competition not in ["E", "U"]:
            raise ValueError(
                "Invalid competition value, {competition}. "
                "Valid values 'E', 'U'"
            )
        self.competition = competition
        self.url_v1 = f"{self.BASE_URL}/{self.V1}/results/"
        self.url_v2 = f"{self.BASE_URL}/{self.V2}/competitions/{competition}"
        # Don't rename url to url_v3, as the former it's used is several places
        self.url = f"{self.BASE_URL}/{self.V3}/competitions/{competition}"

    def make_season_game_url(
        self,
        season: int,
        game_code: int,
        endpoint: str
    ) -> str:
        """
        Concatenates the base URL and makes the game url.

        Args:

            season (int): The start year of the season

            game_code (int): The code of the game. Find the code from
                Euroleague's website

            endpoint (str): The endpoint of the API

        Returns:

            str: the full URL
        """
        full_url = (
            f"{self.url}/seasons/{self.competition}{season}/"
            f"games/{game_code}/{endpoint}"
        )
        return full_url

    def get_gamecodes_season(self, season: int) -> pd.DataFrame:
        """
        A function that returns the game metadata, e.g. gamecodes of season

        Args:

            season (int): The start year of the season.

        Returns:

            pd.DataFrame: A dataframe with the season's game metadata, e.g.
                gamecode, score, home-away teams, date, round, etc.
        """
        params = {
            "seasonCode": f"{self.competition}{season}",
        }
        r = get_requests(self.url_v1, params=params)

        data = xmltodict.parse(r.content)
        df = pd.DataFrame(data["results"]["game"])
        df.rename(
            columns={
                "gamenumber": "gameCode",
                "round": "Phase",
                "gameday": "Round"
            },
            inplace=True
        )
        int_cols = ["Round", "gameCode", "homescore", "awayscore"]
        df[int_cols] = df[int_cols].astype(int)
        df["played"] = df["played"].astype(
            bool).replace({"true": True, "false": False})
        df.sort_values(["gameCode"], ignore_index=True, inplace=True)
        return df

    def get_gamecodes_round(
            self, season: int,
            round_number: int
    ) -> pd.DataFrame:
        """
        A function that returns the game metadata, e.g. gamecodes of a round
        in a season.

        Args:

            season (int): The start year of the season.

            round_number (int): The round number.

        Returns:

            pd.DataFrame: A dataframe with the round_number's game metadata,
                e.g. gamecode, score, home-away teams, date, etc.
        """
        url = f"{self.url_v2}/seasons/{self.competition}{season}/games"
        params = {"roundNumber": round_number}
        r = get_requests(url, params=params)
        try:
            data = r.json()
        except JSONDecodeError as exc:
            raise ValueError(
                f"Round, {round_number}, season {season}, "
                "did not return any data."
            ) from exc

        df = pd.json_normalize(data["data"])
        df.rename(
            columns={
                "round": "Round",
                "phaseType.code": "Phase"
            },
            inplace=True
        )
        df.sort_values(["gameCode"], ignore_index=True, inplace=True)
        return df

    def get_round_data_from_game_data(
        self,
        season: int,
        round_number: int,
        fun: Callable[[int, int], pd.DataFrame]
    ) -> pd.DataFrame:
        """A wrapper function for getting game data for all games in a single
        round.

        Args:
            season (int, optional): The start year of the season.

            round_number (int): The round of the season.

            fun (Callable[[int, int], pd.DataFrame]): A callable function that
                determines that type of data to be collected. Available values:
                - get_game_report
                - get_game_stats
                - get_game_teams_comparison
                - get_game_play_by_play_data
                - get_game_shot_data
                - get_game_boxscore_quarter_data
                - get_player_boxscore_stats_data
                - get_game_metadata


        Returns:
            pd.DataFrame: A dataframe with the corresponding data of a single
                round
        """
        game_codes_df = self.get_gamecodes_round(season, round_number)
        game_codes_df = game_codes_df[game_codes_df["played"]]
        df = get_data_over_collection_of_games(
            game_codes_df,
            season=season,
            fun=fun
        )
        return df

    def get_season_data_from_game_data(
        self,
        season: int,
        fun: Callable[[int, int], pd.DataFrame]
    ) -> pd.DataFrame:
        """
        A wrapper function for getting game data for all games in a single
        season.

        Args:

            season (int, optional): The start year of the season.

            fun (Callable[[int, int], pd.DataFrame]): A callable function that
                determines that type of data to be collected. Available values:
                - get_game_report
                - get_game_stats
                - get_game_teams_comparison
                - get_game_play_by_play_data
                - get_game_shot_data
                - get_game_boxscore_quarter_data
                - get_player_boxscore_stats_data
                - get_game_metadata

        Returns:

            pd.DataFrame: A dataframe with the corresponding data of all
                games in a single season.
        """
        game_codes_df = self.get_gamecodes_season(season)
        game_codes_df = game_codes_df[game_codes_df["played"]]
        season_game_codes_df = (
            game_codes_df[["Phase", "Round", "gameCode"]]
            .drop_duplicates().sort_values(["gameCode", "Round"])
            .reset_index(drop=True)
        )
        df = get_data_over_collection_of_games(
            season_game_codes_df,
            season=season,
            fun=fun
        )
        return df

    def get_range_seasons_data(
        self,
        start_season: int,
        end_season: int,
        fun: Callable[[int, int], pd.DataFrame]
    ) -> pd.DataFrame:
        """
        A wrapper function with the all game data in a range of seasons

        Args:

            season (int, optional): The start year of the season.

            end_season (int): The start year of teh end season

            fun (Callable[[int, int], pd.DataFrame]): A callable function that
                determines that type of data to be collected. Available values:
                - get_game_report
                - get_game_stats
                - get_game_teams_comparison
                - get_game_play_by_play_data
                - get_game_shot_data
                - get_game_boxscore_quarter_data
                - get_player_boxscore_stats_data
                - get_game_metadata

        Returns:

            pd.DataFrame: A dataframe with the corresponding data of all
                games in a range of seasons.
        """
        data = []
        for season in trange(
                start_season, end_season + 1, desc="Season loop", leave=True):

            data_df = self.get_season_data_from_game_data(season, fun)
            data.append(data_df)
        df = pd.concat(data)
        df.reset_index(drop=True, inplace=True)
        return df



import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .utils import get_requests


class Standings(EuroLeagueData):
    """
    A class for getting standings data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_standings(
        self,
        season: int,
        round_number: int,
        endpoint: str = "basicstandings",
    ) -> pd.DataFrame:
        """
        Get the standings of round in given season

        Args:

            season (int): The start year of the season

            round_number (int): The round number

            endpoint (str, optional): The type of standing.
            One of the following options
            - calendarstandings
            - streaks
            - aheadbehind
            - margins
            - basicstandings
            Defaults to "basicstandings".

        Raises:

            ValueError: If endpoint is not applicable

        Returns:

            pd.DataFrame: A dataframe with the standings of the teams
        """
        available_endpoints = [
            "calendarstandings", "streaks",
            "aheadbehind", "margins",
            "basicstandings"
        ]

        if endpoint not in available_endpoints:
            raise ValueError(
                "Standings endpoint, {endpoint}, is not applicable. Choose "
                f"one of the following: {available_endpoints}"
            )

        url_ = (
            f"{self.url}/seasons/{self.competition}{season}/"
            f"rounds/{round_number}/{endpoint}"
        )
        r = get_requests(url_)
        data = r.json()
        df = pd.json_normalize(data["teams"])
        return df
